#
# See the file "LICENSE" for the full license governing this code.
#
newproperty(:permissions, :array_matching => :all) do
  include EasyType
  include EasyType::Mungers::Upcase
  include Puppet_X::EnterpriseModules::Oracle::Information
  include Puppet_X::EnterpriseModules::Oracle::Access

  desc <<-EOD
    The permissions on the specified object, granted to the specified user.

    This is a required array of rights to grant on the user object combination

        ora_object_grant{ ...:
          permissions => ['execute', 'select'],
        }

    When this list contains less rights then currently granted, the the extra rights will
    be revoked from the user. If the permissions list contains more rights then currently
    granted in the database, the extra rights will be granted to the user.

    If you want to make sure no rights are granted, you must use an empty array.

        ora_object_grant{ ...:
          permissions => ['execute', 'select'],
        }

  EOD

  isrequired

  to_translate_to_resource do |raw_resource|
    raw_resource['PRIVS'].split(';')
  end

  def insync?(is)
    if resource.wildcard_table?
      wildcard_insync?
    else
      is = [] if is == :absent || is.nil?
      is.sort == munged_should.sort
    end
  end

  def change_to_s(_from, _to)
    return_value = []
    if resource.wildcard_table?
      grants_to_sync.each do | table|
        current_rights = table['PRIVS'].split(';')
        table_name = table['TABLE_NAME']
        return_value << "revoked the #{revoked_rights(current_rights).join(',')} right(s) on table #{table_name}" unless revoked_rights(current_rights).empty?
        return_value << "granted the #{granted_rights(current_rights).join(',')} right(s) on table #{table_name}" unless granted_rights(current_rights).empty?
      end
    else
      return_value << "revoked the #{revoked_rights.join(',')} right(s)" unless revoked_rights.empty?
      return_value << "granted the #{granted_rights.join(',')} right(s)" unless granted_rights.empty?
    end
    return_value.join(' and ')
  end

  def is_to_s(_)
    super if !resource.wildcard_table?
    messages = []
    grants_to_sync.each do | table|
      current_rights = table['PRIVS'].split(';')
      table_name = table['TABLE_NAME']
      messages << "#{current_rights.inspect} on table #{table_name}"
    end
    messages.join(' and ')
  end

  def should_to_s(should)
    super unless resource.wildcard_table?
    tables =  grants_to_sync.collect { |t| t['TABLE_NAME'] }
    "#{should} on tables #{tables.join(', ')}"
  end

  on_apply do
    if resource.wildcard_table?
      sql(template('ora_config/ora_object_grant/wildcard_update_grants.sql.erb', binding), :sid => resource.sid)
    else
      revoked_rights.each do |right|
        sql(revoke(right), :sid => resource.sid)
      end
      granted_rights.each do |right|
        sql(grant(right), :sid => resource.sid)
      end
    end
    nil
  end

  private

  def wildcard_insync?
    grants_to_sync.empty?
  end

  def grants_to_sync
    unless @grants_to_sync 
      @grants_to_sync = merge_grants(ungranted_tables, actual_grants)
      @grants_to_sync = @grants_to_sync.select { |g| g['PRIVS'].split(';').sort != value }
    end
    @grants_to_sync
  end

  def munged_should
    (should == :absent || should.nil?) ? [] : should
  end

  def current_rights
    # TODO: Check why this needs to be so difficult
    resource.to_resource.to_hash.delete_if { |_key, value| value == :absent }.fetch(:permissions) { [] }
  end

  def expected_rights
    resource.to_hash.fetch(:permissions) { [] }
  end

  def revoked_rights(rights = current_rights)
    rights - expected_rights
  end

  def granted_rights(rights = current_rights )
    expected_rights - rights
  end

  def scope
    case resource[:container]
    when :current
      'CONTAINER = CURRENT'
    when :all
      'CONTAINER = ALL'
    else
      ''
    end
  end

  def revoke(right, object = resource.object_name)
    "revoke #{right} on #{object} from #{resource.grantee} #{scope}"
  end

  def grant(right, object = resource.object_name)
    "grant #{right} on #{object} to #{resource.grantee} #{scope}"
  end

  # 
  # Return Array with information describing ungranted tbales
  #
  def ungranted_tables
    wildcard_tables.map do |table_name|
      table = {}
      table['TABLE_NAME'] = table_name
      table['GRANTEE']    = resource.grantee
      table['PRIVS']      = ''
      table['OWNER']      = resource.object_name.split('.').first
      table
    end
  end

  def actual_grants
    statement = for_version(
      :'12'     => template('ora_config/ora_object_grant/wildcard_index_v12.sql.erb', binding),
      :default  => template('ora_config/ora_object_grant/wildcard_index.sql.erb', binding)
    )
    sql(statement, :sid => resource.sid)
  end

  def merge_grants(base, merge)
    merge.each do | grant|
      table_name = grant['TABLE_NAME']
      record = base.find {|r| r['TABLE_NAME'] == table_name}
      next if record.nil?
      record['PRIVS'] = grant['PRIVS']
    end
    base
  end

  def wildcard_tables
    sql(template('ora_config/ora_object_grant/get_tables.sql.erb', binding), :sid => resource.sid).collect {|r| r['OBJECT_NAME']} 
  end
end